import unittest
import os
import fernet_files
from fernet_files.custom_fernet import FernetNoBase64
from io import BytesIO, UnsupportedOperation
from random import randint
from typing import Callable
try:
    from tqdm import tqdm # optional progress bar
    TQDM_AVAILABLE = True
except ImportError:
    TQDM_AVAILABLE = False

# Variables:
# BytesIO or normal file
# with or nowith
# invalid, valid, valid from cryptography, incorrect key
# read-only, or read-write file
# chunksize: different, invalid
# reading different sizes
# input data: smaller than, equal to and larger than chunk size, empty
# input data: see if random or specific cases cause issues
# test different seeking offsets: random, invalid, smaller, equal and larger than chunk size
# test different whences

# must test EVERY combination of these variables to pass, don't miss anything!

def chunk_testing_sizes():
    yield 1 # 1B
    yield 256 # 256B
    yield 65536 # 64KiB
    yield 1_048_576 # 1MiB
    for _ in range(3):
        yield randint(1, 100_000) # 1B to 100KB
    for _ in range(3):
        yield randint(100_000, 10_000_000) # 100KB to 10MB

def input_testing_sizes(chunksize):
    yield 0
    yield 1
    yield chunksize
    yield chunksize + 73 - (chunksize % 16)
    yield chunksize*100
    yield chunksize*100 + 73 - (chunksize*100 % 16)

def execute_test(desc: str, test: Callable) -> None:
    if TQDM_AVAILABLE:
        pbar = tqdm(desc=desc, total=60) # magic, must be changing if number of test sizes changes
    for chunksize in chunk_testing_sizes():
        for inputsize in input_testing_sizes(chunksize):
            input_data = os.urandom(inputsize)
            test(chunksize, input_data)
            if TQDM_AVAILABLE: pbar.update(1)

class TestFernetFiles(unittest.TestCase):
    def test_invalid_chunksizes(self):
        for chunksize in (1.5, "1", b"1"):
            self.assertRaises(TypeError, fernet_files.FernetFile, fernet_files.FernetFile.generate_key(), BytesIO(), chunksize=chunksize)
        for chunksize in (0, -1, -2):
            self.assertRaises(ValueError, fernet_files.FernetFile, fernet_files.FernetFile.generate_key(), BytesIO(), chunksize=chunksize)

    def test_key(self):
        # test generate key
        self.assertEqual(fernet_files.FernetFile.generate_key, FernetNoBase64.generate_key)
        key = fernet_files.FernetFile.generate_key()
        self.assertIsInstance(key, bytes)
        self.assertEqual(len(key), 32)
        f = FernetNoBase64(key) # valid key
        with fernet_files.FernetFile(FernetNoBase64.generate_key(), BytesIO()) as fernet_file: pass
        self.assertRaises(ValueError, fernet_files.FernetFile, os.urandom(33), BytesIO()) # invalid key
        self.assertRaises(TypeError, fernet_files.FernetFile, int.from_bytes(os.urandom(32), "little"), BytesIO())
        # test with FernetNoBase64 object
        with fernet_files.FernetFile(f, BytesIO()) as fernet_file: pass

    def test_invalid_file(self):
        for chunksize in chunk_testing_sizes():
            with open("test", "wt") as invalid_file:
                self.assertRaises(TypeError, fernet_files.FernetFile, fernet_files.FernetFile.generate_key(), invalid_file, chunksize=chunksize)

    def test_file_nowith_readwrite(self):
        def test(chunksize, input_data):
            key = fernet_files.FernetFile.generate_key()
            with open("test", "wb+") as f:
                fernet_file = fernet_files.FernetFile(key, f, chunksize)
                fernet_file.write(input_data)
                fernet_file.seek(0)
                test_seeking(self, fernet_file, chunksize, input_data)
                test_random_reads(self, fernet_file, chunksize, input_data)
                test_other_read(self, fernet_file, input_data)
                input_data = test_random_writes(self, fernet_file, chunksize, input_data)
                test_data_is_encrypted(self, "test", fernet_file, input_data)
                fernet_file.close()
                self.assertRaises(ValueError, fernet_file.seek, 0)
                self.assertRaises(ValueError, fernet_file.read)
                self.assertRaises(ValueError, fernet_file.write, input_data)
            test_random_write_withclose(self, "test", chunksize, input_data)
        execute_test("test_file_nowith_readwrite", test)

    def test_file_nowith_readonly(self):
        def test(chunksize, input_data):
            key = fernet_files.FernetFile.generate_key()
            with open("test", "wb+") as f:
                fernet_file = fernet_files.FernetFile(key, f, chunksize)
                fernet_file.write(input_data)
                fernet_file.close()
            with open("test", "rb") as f:
                fernet_file = fernet_files.FernetFile(key, f, chunksize)
                self.assertRaises(UnsupportedOperation, fernet_file.write, input_data)
                fernet_file.seek(0)
                test_seeking(self, fernet_file, chunksize, input_data)
                test_random_reads(self, fernet_file, chunksize, input_data)
                test_other_read(self, fernet_file, input_data)
                test_data_is_encrypted(self, "test", fernet_file, input_data)
                fernet_file.close()
                self.assertRaises(ValueError, fernet_file.seek, 0)
                self.assertRaises(ValueError, fernet_file.read)
        execute_test("test_file_nowith_readonly", test)

    def test_bytesio_nowith(self):
        def test(chunksize, input_data):
            key = fernet_files.FernetFile.generate_key()
            with BytesIO() as f:
                fernet_file = fernet_files.FernetFile(key, f, chunksize)
                fernet_file.write(input_data)
                fernet_file.seek(0)
                test_seeking_bytesio(self, fernet_file, chunksize, input_data)
                test_random_reads(self, fernet_file, chunksize, input_data)
                test_other_read(self, fernet_file, input_data)
                input_data = test_random_writes(self, fernet_file, chunksize, input_data)
                test_data_is_encrypted(self, f, fernet_file, input_data)
                fernet_file.close()
                self.assertRaises(ValueError, fernet_file.seek, 0)
                self.assertRaises(ValueError, fernet_file.read)
                self.assertRaises(ValueError, fernet_file.write, input_data)
        execute_test("bytesio_nowith", test)

    def test_file_with_readwrite(self):
        def test(chunksize, input_data):
            key = fernet_files.FernetFile.generate_key()
            with open("test", "wb+") as f:
                with fernet_files.FernetFile(key, f, chunksize) as fernet_file:
                    fernet_file.write(input_data)
                    fernet_file.seek(0)
                    test_seeking(self, fernet_file, chunksize, input_data)
                    test_random_reads(self, fernet_file, chunksize, input_data)
                    test_other_read(self, fernet_file, input_data)
                    input_data = test_random_writes(self, fernet_file, chunksize, input_data)
                self.assertRaises(ValueError, fernet_file.seek, 0)
                self.assertRaises(ValueError, fernet_file.read)
                self.assertRaises(ValueError, fernet_file.write, input_data)
            test_random_write_withclose_withwith(self, "test", chunksize, input_data)
        execute_test("test_file_with_readwrite", test)

    def test_file_with_readonly(self):
        def test(chunksize, input_data):
            key = fernet_files.FernetFile.generate_key()
            with open("test", "wb+") as f:
                with fernet_files.FernetFile(key, f, chunksize) as fernet_file:
                    fernet_file.write(input_data)
            with open("test", "rb") as f:
                with fernet_files.FernetFile(key, f, chunksize) as fernet_file:
                    self.assertRaises(UnsupportedOperation, fernet_file.write, input_data)
                    fernet_file.seek(0)
                    test_seeking(self, fernet_file, chunksize, input_data)
                    test_random_reads(self, fernet_file, chunksize, input_data)
                    test_other_read(self, fernet_file, input_data)
                    test_data_is_encrypted(self, "test", fernet_file, input_data)
                self.assertRaises(ValueError, fernet_file.seek, 0)
                self.assertRaises(ValueError, fernet_file.read)
        execute_test("test_file_with_readonly", test)

    def test_bytesio_with(self):
        def test(chunksize, input_data):
            key = fernet_files.FernetFile.generate_key()
            with BytesIO() as f:
                with fernet_files.FernetFile(key, f, chunksize) as fernet_file:
                    fernet_file.write(input_data)
                    fernet_file.seek(0)
                    test_seeking_bytesio(self, fernet_file, chunksize, input_data)
                    test_random_reads(self, fernet_file, chunksize, input_data)
                    test_other_read(self, fernet_file, input_data)
                    input_data = test_random_writes(self, fernet_file, chunksize, input_data)
                    test_data_is_encrypted(self, f, fernet_file, input_data)
                self.assertRaises(ValueError, fernet_file.seek, 0)
                self.assertRaises(ValueError, fernet_file.read)
                self.assertRaises(ValueError, fernet_file.write, input_data)
        execute_test("bytesio_with", test)

def test_seeking(unit_test: TestFernetFiles, fernet_file: fernet_files.FernetFile, chunksize: int, input_data: bytes) -> None:
    for get_size in (lambda: randint(0, chunksize-1), lambda: chunksize, lambda: randint(chunksize+1, chunksize*3)): # below, equal, above chunksize
        for x in (randint(0, len(input_data)-1 if input_data else 0) for _ in range(100)): # random starting points
            size = get_size()
            data = input_data[x:x+size]
            unit_test.assertEqual(fernet_file.seek(x), x) # variations of whence input
            unit_test.assertEqual(fernet_file.read(size), data)
            unit_test.assertEqual(fernet_file.seek(x, os.SEEK_SET), x)
            unit_test.assertEqual(fernet_file.read(size), data)
            unit_test.assertEqual(fernet_file.seek(x, 0), x)
            unit_test.assertEqual(fernet_file.read(size), data)
        last = fernet_file.seek(0)
        for _ in range(100):
            size = get_size()
            x = randint(-last, len(input_data)-last-1 if input_data else 0)
            data = input_data[last+x:last+x+size]
            unit_test.assertEqual(fernet_file.seek(x, os.SEEK_CUR), last+x)
            unit_test.assertEqual(fernet_file.read(size), data)
            unit_test.assertEqual(fernet_file.seek(-x-len(data), os.SEEK_CUR), last)
            unit_test.assertEqual(fernet_file.seek(x, 1), last+x)
            unit_test.assertEqual(fernet_file.read(size), data)
            last += x+len(data)
        for x in (randint(-len(input_data)+1 if input_data else 0, 0) for _ in range(100)):
            size = get_size()
            if x+size < 0:
                data = input_data[x:x+size]
            elif x == 0:
                data = b''
            else:
                data = input_data[x:]
            unit_test.assertEqual(fernet_file.seek(x, os.SEEK_END), len(input_data)+x)
            unit_test.assertEqual(fernet_file.read(size), data)
            unit_test.assertEqual(fernet_file.seek(x, 2), len(input_data)+x)
            unit_test.assertEqual(fernet_file.read(size), data)
    unit_test.assertRaises(OSError, fernet_file.seek, -1) # Negative
    unit_test.assertRaises(ValueError, fernet_file.seek, 0, 3) # Invalid whence

def test_seeking_noread(unit_test: TestFernetFiles, fernet_file: fernet_files.FernetFile, chunksize: int, input_data: bytes) -> None:
    for x in (randint(0, len(input_data)-1 if input_data else 0) for _ in range(100)): # random starting points
        unit_test.assertEqual(fernet_file.seek(x), x) # variations of whence input
        unit_test.assertEqual(fernet_file.seek(x, os.SEEK_SET), x)
        unit_test.assertEqual(fernet_file.seek(x, 0), x)
    last = fernet_file.seek(0)
    for _ in range(100):
        x = randint(-last, len(input_data)-last-1 if input_data else 0)
        unit_test.assertEqual(fernet_file.seek(x, os.SEEK_CUR), last+x)
        unit_test.assertEqual(fernet_file.seek(-x, os.SEEK_CUR), last)
        unit_test.assertEqual(fernet_file.seek(x, 1), last+x)
        last += x
    for x in (randint(-len(input_data)+1 if input_data else 0, 0) for _ in range(100)):
        unit_test.assertEqual(fernet_file.seek(x, os.SEEK_END), len(input_data)+x)
        unit_test.assertEqual(fernet_file.seek(x, 2), len(input_data)+x)
    unit_test.assertRaises(OSError, fernet_file.seek, -1) # Negative
    unit_test.assertRaises(ValueError, fernet_file.seek, 0, 3) # Invalid whence

def test_seeking_bytesio(unit_test: TestFernetFiles, fernet_file: fernet_files.FernetFile, chunksize: int, input_data: bytes) -> None:
    for get_size in (lambda: randint(0, chunksize-1), lambda: chunksize, lambda: randint(chunksize+1, chunksize*3)): # below, equal, above chunksize
        for x in (randint(0, len(input_data)-1 if input_data else 0) for _ in range(100)): # random starting points
            size = get_size()
            data = input_data[x:x+size]
            unit_test.assertEqual(fernet_file.seek(x), x)
            y = fernet_file.read(size)
            unit_test.assertEqual(y, data)
            unit_test.assertRaises(ValueError, fernet_file.seek, x, -1) # ignored whence
    unit_test.assertRaises(ValueError, fernet_file.seek, -1) # Negative

def test_random_reads(unit_test: TestFernetFiles, fernet_file: fernet_files.FernetFile, chunksize: int, input_data: bytes) -> None:
    for _ in range(100): # Read data below chunksize
        x = randint(0, readrange) if (readrange := len(input_data)-chunksize-1) >= 0 else 0
        y = randint(x, x+chunksize)
        fernet_file.seek(x)
        unit_test.assertEqual(fernet_file.read(y-x), input_data[x:y])
    for _ in range(100): # Read data equal to chunksize
        x = randint(0, readrange) if (readrange := len(input_data)-chunksize-1) >= 0 else 0
        fernet_file.seek(x)
        unit_test.assertEqual(fernet_file.read(chunksize), input_data[x:x+chunksize])
    for _ in range(100): # Read data above chunksize
        x = randint(0, readrange) if (readrange := len(input_data)-chunksize-1) >= 0 else 0
        y = randint(x+chunksize+1, x+chunksize*3)
        fernet_file.seek(x)
        unit_test.assertEqual(fernet_file.read(y-x), input_data[x:y])

def test_other_read(unit_test: TestFernetFiles, fernet_file: fernet_files.FernetFile, input_data: bytes) -> None:
    fernet_file.seek(0)
    unit_test.assertEqual(fernet_file.read(), input_data) # Full
    fernet_file.seek(0)
    unit_test.assertEqual(fernet_file.read(-1), input_data) # Negative
    fernet_file.seek(0)
    unit_test.assertEqual(fernet_file.read(0), b"") # Empty
    unit_test.assertRaises(TypeError, fernet_file.read, 1.5) # Float
    unit_test.assertRaises(TypeError, fernet_file.read, "1") # String

def test_random_writes_noread(fernet_file: fernet_files.FernetFile, chunksize: int, input_data: bytes) -> bytes:
    input_data = BytesIO(input_data)
    for get_size in (lambda: randint(0, chunksize-1), lambda: chunksize, lambda: randint(chunksize+1, chunksize*3)): # below, equal, above chunksize
        for _ in range(100): # Get random size and location and write there in file and input data
            size = get_size()
            input_data.seek(0)
            length = len(input_data.read())
            x = randint(0, length-1 if length else 0)
            randata = os.urandom(size)
            input_data.seek(x)
            input_data.write(randata)
            fernet_file.seek(x)
            fernet_file.write(randata)
    return input_data.getvalue()

def test_random_writes(unit_test: TestFernetFiles, fernet_file: fernet_files.FernetFile, chunksize: int, input_data: bytes) -> bytes:
    input_data = test_random_writes_noread(fernet_file, chunksize, input_data)
    fernet_file.seek(0)
    unit_test.assertEqual(fernet_file.read(), input_data)
    return input_data

def test_random_write_withclose(unit_test: TestFernetFiles, filename: str, chunksize: int, input_data: bytes) -> None:
    key = fernet_files.FernetFile.generate_key()
    with open(filename, "wb+") as f:
        fernet_file = fernet_files.FernetFile(key, f)
        fernet_file.write(input_data)
        input_data = test_random_writes(unit_test, fernet_file, chunksize, input_data)
        fernet_file.close()
    with open(filename, "rb") as f:
        fernet_file = fernet_files.FernetFile(key, f)
        unit_test.assertEqual(fernet_file.read(), input_data)
        fernet_file.close()

def test_random_write_withclose_withwith(unit_test: TestFernetFiles, filename: str, chunksize: int, input_data: bytes) -> None:
    key = fernet_files.FernetFile.generate_key()
    with open(filename, "wb+") as f:
        with fernet_files.FernetFile(key, f) as fernet_file:
            fernet_file.write(input_data)
            input_data = test_random_writes(unit_test, fernet_file, chunksize, input_data)
    with open(filename, "rb") as f:
        with fernet_files.FernetFile(key, f) as fernet_file:
            unit_test.assertEqual(fernet_file.read(), input_data)

def test_data_is_encrypted(unit_test: TestFernetFiles, file: str | BytesIO, fernet_file: fernet_files.FernetFile, input_data: bytes) -> None:
    if isinstance(file, str):
        with open(file, "rb") as f:
            output_data = f.read() # Encrypted data
    else:
        file.seek(0)
        output_data = file.read()
    fernet_file.seek(0)
    unit_test.assertNotEqual(input_data, output_data)
                
if __name__ == '__main__':
    unittest.main()
    os.remove("test")


from fernet_files.custom_fernet import FernetNoBase64
import os
import os.path
from io import BytesIO, RawIOBase, BufferedIOBase, StringIO, TextIOBase, UnsupportedOperation

# Don't modify without reading documentation
META_SIZE = 8

DEFAULT_CHUNKSIZE = 65536

class FernetFile:
    def __init__(self, key: bytes | FernetNoBase64, file: str | RawIOBase | BufferedIOBase, chunksize: int = DEFAULT_CHUNKSIZE) -> None:
        self.closed = False

        self.__fernet = key if isinstance(key, FernetNoBase64) else FernetNoBase64(key) # key validation
        # file validation
        if isinstance(file, (StringIO, TextIOBase)):
            raise TypeError("File provided must be binary, not string")
        elif isinstance(file, (RawIOBase, BufferedIOBase, BytesIO)):
            self.__file = file
        elif isinstance(file, str):
            if os.path.exists(file):
                self.__file = open(file, "rb+")
            else:
                self.__file = open(file, "wb+")
        else:
            raise TypeError("File must be binary file or a filename")
        
        # chunksize validation
        if not isinstance(chunksize, int):
            raise TypeError("Invalid chunksize, must be integer greater than 0")
        if chunksize <= 0:
            raise ValueError("Invalid chunksize, must be integer greater than 0")

        # get metadata
        self.__file.seek(0)
        if x := self.__file.read(META_SIZE): # If metadata exists, read it
            self.__last_chunk = int.from_bytes(x, "little")
            self.__last_chunk_padding = int.from_bytes(self.__file.read(META_SIZE), "little")
        else:
            self.__last_chunk, self.__last_chunk_padding = 0, 0
        # write metadata + check writeability
        self.__file.seek(0)
        try:
            self.__file.write(bytes(META_SIZE*2))
            self.writeable = True
        except UnsupportedOperation:
            self.writeable = False

        self.__data_chunksize = chunksize # the size of the data in chunks
        self.__chunksize = chunksize + 73 - (chunksize % 16) # the size of chunks when written to disk
        
        self.__chunk_modified = False
        self._pos_pointer = 0 # your position inside a chunk
        self._chunk_pointer = 0 # what chunk you're currently in

    def __goto_current_chunk(self) -> None:
        self.__file.seek(self._chunk_pointer*self.__chunksize+META_SIZE*2)

    def __get_file_size(self) -> int:
        return (self.__last_chunk+1)*self.__data_chunksize-self.__last_chunk_padding
    
    def __read_chunk(self) -> BytesIO:
        if self.__chunk_modified:
            return self.__chunk
            # you can't modify a chunk without it already being loaded
        self.__goto_current_chunk()
        try:
            data = self.__fernet.decrypt(self.__file.read(self.__chunksize))
            if self._chunk_pointer == self.__last_chunk and self.__last_chunk_padding:
                data = data[:-self.__last_chunk_padding]
            self.__chunk = BytesIO(data)
        except:
            self.__chunk = BytesIO()
        return self.__chunk
    
    def __write_chunk(self) -> None:
        if not self.writeable:
            return # Raising an exception is the write method's responsibility
        self.__goto_current_chunk()
        data = self.__chunk.getvalue()
        padding = self.__data_chunksize - len(data)
        data += bytes(padding)
        self.__file.write(self.__fernet.encrypt(data))
        if self._chunk_pointer >= self.__last_chunk:
            self.__last_chunk = self._chunk_pointer
            self.__last_chunk_padding = padding
            self.__file.seek(0)
            self.__file.write(self.__last_chunk.to_bytes(META_SIZE, "little"))
            self.__file.write(self.__last_chunk_padding.to_bytes(META_SIZE, "little"))
        self.__chunk_modified = False

    def seek(self, *args, whence: int = os.SEEK_SET) -> int:
        if self.closed:
            raise ValueError("I/O operation on closed file")

        # argument parsing
        if len(args) >= 2:
            offset, whence = args[:2]
        else:
            offset = args[0]

        original = self._chunk_pointer, self._pos_pointer

        try:
            if whence == os.SEEK_SET or whence == 0:
                self._chunk_pointer = 0
                self._pos_pointer = offset
            elif whence == os.SEEK_CUR or whence == 1:
                self._pos_pointer += offset
            elif whence == os.SEEK_END or whence == 2:
                size = self.__get_file_size()
                self._chunk_pointer = 0
                self._pos_pointer = size + offset
            else:
                raise ValueError("Invalid whence")
        except OSError:
            self._chunk_pointer, self._pos_pointer = original
            raise OSError("Invalid seek value")

        return self._pos_pointer + self._chunk_pointer*self.__data_chunksize

    def read(self, size: int = -1) -> bytes:
        if self.closed:
            raise ValueError("I/O operation on closed file")
        # data validation
        if not isinstance(size, int):
            raise TypeError("Size must be an integer")
        if size < 0:
            self.__write_chunk() # refreshes values for last chunk
            size = self.__get_file_size() - self._pos_pointer - self._chunk_pointer*self.__data_chunksize

        self.__read_chunk()
        if size <= self.__data_chunksize-self._pos_pointer: # if all wanted data is in current chunk
            self.__chunk.seek(self._pos_pointer)
            data = self.__chunk.read(size)
            self._pos_pointer += size
        else: # else: read until start of next chunk
            # read in the current chunk
            data = bytearray()
            read_size = self.__data_chunksize-self._pos_pointer
            self.__chunk.seek(self._pos_pointer)
            data += self.__chunk.read(read_size)
            self._pos_pointer = 0
            self._chunk_pointer += 1
            size -= read_size
            # loop, reading all the data in each chunk
            while size >= self.__data_chunksize:
                data += self.__read_chunk().getvalue()
                size -= self.__data_chunksize
                self._chunk_pointer += 1
            # partially read the last chunk
            if size:
                data += self.__read_chunk().read(size)
                self._pos_pointer = size

        # For some reason if you read to the end of a file in python, the pointer should be at the end of the file
        # Original implementation puts the pointer at the distance specified by the read argument, this fixes that
        filesize = self.__get_file_size()
        if self._pos_pointer + self._chunk_pointer*self.__data_chunksize >= filesize:
            self.__chunk_pointer = 0
            self._pos_pointer = self.__get_file_size()

        return bytes(data)

    def write(self, b: bytes) -> int:
        if not self.writeable:
            raise UnsupportedOperation("write")
        if self.closed:
            raise ValueError("I/O operation on closed file")
        # data validation
        if not isinstance(b, bytes):
            raise TypeError("Data must be bytes")
        size = len(b)
        return_value = size

        if size < self.__data_chunksize-self._pos_pointer: # if all data fits in current chunk
            if self.__chunk is not None:
                self.__chunk.seek(self._pos_pointer)
                self.__chunk.write(b)
            else:
                self.__chunk = BytesIO(b)
            self._pos_pointer += size
            self.__chunk_modified = True
            return size
        # else: write until start of next chunk. else omitted for indent readability
        # write in the current chunk
        b = BytesIO(b)
        write_size = self.__data_chunksize-self._pos_pointer
        if self.__chunk is not None:
            self.__chunk.seek(self._pos_pointer)
            self.__chunk.write(b.read(write_size))
        else:
            self.__chunk = BytesIO(b.read(write_size))
        self._pos_pointer = 0
        self.__chunk_modified = True
        size -= write_size
        # loop, writing full chunks
        while size >= self.__data_chunksize:
            self._chunk_pointer += 1
            self.__chunk = BytesIO(b.read(self.__data_chunksize))
            self.__chunk_modified = True
            size -= self.__data_chunksize
        # partially write the last chunk
        if size:
            self._chunk_pointer += 1
            self.__chunk = self.__read_chunk()
            self.__chunk.seek(0)
            self.__chunk.write(b.read(size))
            self._pos_pointer = size
            self.__chunk_modified = True
        return return_value
    
    def close(self) -> BytesIO | None:
        # write data stored in memory
        try: self.__write_chunk()
        except: pass
        # mark as closed
        self.closed = True
        try:
            # if file is BytesIO, return it, otherwise close the file
            if isinstance(self.__file, BytesIO):
                return self.__file
            else:
                try: self.__file.close()
                except: pass
        except AttributeError:
            # fixes https://github.com/Kris-0605/fernet_files/issues/4
            pass

    generate_key = FernetNoBase64.generate_key

    def __enter__(self) -> "FernetFile":
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    @property
    def _pos_pointer(self) -> int:
        return self.__pos_pointer
    
    @_pos_pointer.setter
    def _pos_pointer(self, value: int) -> None:
        self.__pos_pointer = value
        # wrapping value of pos pointer by changing chunk pointer
        if self.__pos_pointer >= self.__data_chunksize:
            chunk_difference = self.__pos_pointer//self.__data_chunksize
            self.__pos_pointer -= chunk_difference*self.__data_chunksize
            self._chunk_pointer += chunk_difference
        elif self.__pos_pointer < 0:
            chunk_difference = -(self.__pos_pointer//self.__data_chunksize)
            self.__pos_pointer += chunk_difference*self.__data_chunksize
            self._chunk_pointer -= chunk_difference
            if self._chunk_pointer < 0:
                raise OSError("Seek before beginning of file")

    @property
    def _chunk_pointer(self) -> int:
        return self.__chunk_pointer
    
    @_chunk_pointer.setter
    def _chunk_pointer(self, value: int) -> None:
        # write a chunk only if it's been modified
        if self.__chunk_modified: self.__write_chunk()
        self.__chunk_pointer = value
        self.__read_chunk()


# This file contains a modified version of code found within cryptography.fernet.Fernet.
# The purpose of this modification is to remove base64 encoding/decoding.
# While base64 is required for the Fernet spec, it adds processing and storage overhead not necessary when storing in files.
# In situations where it is not required, removing base64 from the class increases speed of execution.
#
# This code is based on https://github.com/pyca/cryptography/blob/f558199dbf33ccbf6dce8150c2cd4658686d6018/src/cryptography/fernet.py
# Therefore, the original code is subject to the same licenses as the above link.
# Please see the cryptography library's license at the time of the linked commit for more information.
#
# Any code that I have modifed is labelled with a comment.
# I have removed all comments which I didn't create.

from cryptography.fernet import Fernet, InvalidToken
from cryptography import utils
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.hmac import HMAC
from os import urandom # Changes to imports
from time import time

class FernetNoBase64(Fernet):
    def __init__(
        self,
        key: bytes # Modified to expect raw key, not base64 encoded key
    ) -> None: # Removal of unused parameter backend
        # Modified input validation to match new input
        if len(key) != 32:
            raise ValueError(
                # Modified error messages
                "Fernet key must be 32 bytes"
            )
        if not isinstance(key, bytes):
            raise TypeError(
                "Fernet key must be 32 bytes"
            )

        self._signing_key = key[:16]
        self._encryption_key = key[16:]

    @staticmethod # Conversion from class methdo to static method
    def generate_key() -> bytes:
        return urandom(32) # No base64 encoding

    @staticmethod
    def _get_unverified_token_data(data: bytes) -> tuple[int, bytes]:
        # Modified type hinting so _get_unverified_token_data can't take str
        if not isinstance(data, bytes): # Modified input validation
            raise TypeError("data must be bytes")

        if not data or data[0] != 0x80:
            raise InvalidToken

        if len(data) < 9:
            raise InvalidToken

        timestamp = int.from_bytes(data[1:9], byteorder="big")
        return timestamp, data

    def _encrypt_from_parts(
        self, data: bytes, current_time: int, iv: bytes
    ) -> bytes:
        utils._check_bytes("data", data)

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(data) + padder.finalize()
        encryptor = Cipher(
            algorithms.AES(self._encryption_key),
            modes.CBC(iv),
        ).encryptor()
        ciphertext = encryptor.update(padded_data) + encryptor.finalize()

        basic_parts = (
            b"\x80"
            + current_time.to_bytes(length=8, byteorder="big")
            + iv
            + ciphertext
        )

        h = HMAC(self._signing_key, hashes.SHA256())
        h.update(basic_parts)
        hmac = h.finalize()
        return basic_parts + hmac # Removal of base64 encoding
    
    def decrypt(self, token: bytes, ttl: int | None = None) -> bytes: # Modified type hinting so decrypt can't accept str
        timestamp, data = FernetNoBase64._get_unverified_token_data(token) # Modified which class is referenced
        if ttl is None:
            time_info = None
        else:
            time_info = (ttl, int(time.time()))
        return self._decrypt_data(data, timestamp, time_info)

    def decrypt_at_time(
        self, token: bytes, ttl: int, current_time: int
    ) -> bytes: # Modified type hinting so decrypt_at_time can't take str
        if ttl is None:
            raise ValueError(
                "decrypt_at_time() can only be used with a non-None ttl"
            )
        timestamp, data = FernetNoBase64._get_unverified_token_data(token) # Modified which class is referenced
        return self._decrypt_data(data, timestamp, (ttl, current_time))

    def extract_timestamp(self, token: bytes) -> int: # Modified type hinting so extract_timestamp can't take str
        timestamp, data = FernetNoBase64._get_unverified_token_data(token) # Modified which class is referenced
        self._verify_signature(data)
        return timestamp